from tensorkube.services.eksctl_service import check_and_install_eksctl
from tensorkube.services.istio import check_and_install_istioctl
from tensorkube.services.kubectl_utils import check_and_install_kubectl, check_and_install_helm


def check_and_install_cli_tools():
    # check and install eksctl
    check_and_install_eksctl()
    # check if kubectl is present and if not install
    check_and_install_kubectl()
    # check and install istioctl
    check_and_install_istioctl()
    # check and install helm
    check_and_install_helm()


import platform
import subprocess


def check_and_install_kubectl():
    """Check if kubectl is installed and install it if it's not."""
    try:
        subprocess.run(["kubectl", "version", "--client"], check=True)
        print("kubectl is already installed.")
    except (subprocess.CalledProcessError, FileNotFoundError):
        print("kubectl is not installed. Proceeding with installation.")
        if platform.system() == "Darwin":
            try:
                subprocess.run(["brew", "install", "kubectl"], check=True)
            except subprocess.CalledProcessError as e:
                print("Unable to install kubectl using Homebrew. Please install kubectl manually.")
                raise e
        elif platform.system() == "Linux":
            try:
                with open('/etc/os-release') as f:
                    os_info = f.read().lower()
                    if 'debian' in os_info or 'ubuntu' in os_info:
                        # Check if Snap is installed
                        snap_installed = subprocess.run(["which", "snap"], check=False).returncode == 0

                        if not snap_installed:
                            # Install Snap
                            subprocess.run(["sudo", "apt-get", "update"], check=True)
                            subprocess.run(["sudo", "apt-get", "install", "-y", "snapd"], check=True)

                        # Install kubectl using Snap
                        subprocess.run(["sudo", "snap", "install", "kubectl", "--classic"], check=True)
                    else:
                        print("Unsupported Linux distribution. Please install kubectl manually.")
                        raise Exception('Unsupported Linux distribution.')
            except subprocess.CalledProcessError as e:
                print("Unable to install kubectl. Please install kubectl manually.")
                raise e
        else:
            print("Unsupported operating system. Please install kubectl manually.")
            raise Exception('Unsupported operating system.')

        # Verify kubectl installation
        try:
            subprocess.run(["kubectl", "version", "--client"], check=True)
            print("kubectl installed successfully.")
        except subprocess.CalledProcessError as e:
            print("kubectl installation failed. Please install kubectl manually.")
            raise e


def check_and_install_helm():
    """Check if helm is installed and install it if it's not."""
    try:
        subprocess.run(["helm", "version"], check=True)
        print("Helm is already installed.")
    except (subprocess.CalledProcessError, FileNotFoundError):
        print("Helm is not installed. Proceeding with installation.")
        if platform.system() == "Darwin":
            try:
                subprocess.run(["brew", "install", "helm"], check=True)
            except subprocess.CalledProcessError as e:
                print("Unable to install helm using Homebrew. Please install helm manually.")
                raise e
        elif platform.system() == "Linux":
            try:
                # Check if Snap is installed
                snap_installed = subprocess.run(["which", "snap"], check=False).returncode == 0

                if not snap_installed:
                    # Install Snap
                    distro = platform.linux_distribution()[0].lower()
                    if 'ubuntu' in distro or 'debian' in distro:
                        subprocess.run(["sudo", "apt-get", "update"], check=True)
                        subprocess.run(["sudo", "apt-get", "install", "-y", "snapd"], check=True)
                    elif 'centos' in distro or 'redhat' in distro:
                        subprocess.run(["sudo", "yum", "update"], check=True)
                        subprocess.run(["sudo", "yum", "install", "-y", "epel-release"], check=True)
                        subprocess.run(["sudo", "yum", "install", "-y", "snapd"], check=True)
                        subprocess.run(["sudo", "systemctl", "enable", "--now", "snapd.socket"], check=True)
                        subprocess.run(["sudo", "ln", "-s", "/var/lib/snapd/snap", "/snap"], check=True)
                    else:
                        print("Unsupported Linux distribution. Please install helm manually.")
                        raise Exception('Unsupported Linux distribution.')

                # Install helm using Snap
                subprocess.run(["sudo", "snap", "install", "helm", "--classic"], check=True)
                print("Helm installed successfully using Snap.")
            except subprocess.CalledProcessError as e:
                print("Unable to install helm using Snap. Please install helm manually.")
                raise e
        else:
            print("Unsupported operating system. Please install helm manually.")
            raise Exception('Unsupported operating system.')

        # Verify helm installation
        try:
            subprocess.run(["helm", "version"], check=True)
            print("Helm installed successfully.")
        except subprocess.CalledProcessError as e:
            print("Helm installation failed. Please install helm manually.")
            raise e



import base64
import hashlib
import json
import uuid

import click
import requests

from tensorkube.constants import NAMESPACE, SERVICE_ACCOUNT_NAME, REGION
from tensorkube.services.eks_service import get_cluster_oidc_issuer_url
from tensorkube.services.iam_service import create_s3_csi_driver_role, attach_role_policy


def create_mountpoint_driver_role_with_policy(cluster_name, account_no, role_name, policy_name,
                                              service_account_name=SERVICE_ACCOUNT_NAME, namespace=NAMESPACE,
                                              region=REGION):
    oidc_issuer_url = get_cluster_oidc_issuer_url(cluster_name)
    create_s3_csi_driver_role(account_no, role_name, oidc_issuer_url, namespace, service_account_name)
    attach_role_policy(account_no, policy_name, role_name, region)


def get_base64_encoded_docker_config(username: str, password: str, email: str):
    auth = base64.b64encode(f"{username}:{password}".encode("utf-8")).decode("utf-8")

    docker_config_dict = {"auths": {
        "https://index.docker.io/v1/": {"username": username, "password": password, "email": email, "auth": auth, }}}

    base64_encoded_docker_config = base64.b64encode(json.dumps(docker_config_dict).encode("utf-8")).decode("utf-8")
    return base64_encoded_docker_config


def sanitise_name(name: str):
    return name.replace("_", "-").replace(" ", "-").lower()


def sanitise_assumed_role_arn(arn: str):
    arn = arn.replace('assumed-role', 'role')
    last_slash_index = arn.rfind('/')
    if last_slash_index != -1:
        arn = arn[:last_slash_index]
    return arn


def track_event(event_name: str, event_properties: dict):
    try:
        mac = uuid.getnode()
        id = hashlib.sha256(str(mac).encode())
        id.hexdigest()
        url = 'https://api.tensorfuse.io/tensorfuse/track/tensorkube-event/'

        body = {"id": id.hexdigest(), "event": event_name, "properties": event_properties}
        x = requests.post(url, json=body)
        click.echo(x.text)
    except Exception as e:
        click.echo(f"Error while tracking event: {e}")



import pprint
import threading
import time
from typing import Tuple

import botocore.exceptions
import click
from pkg_resources import resource_filename

from tensorkube.services.aws_service import get_cloudformation_client, get_ec2_client, get_iam_client
from tensorkube.constants import get_cluster_name


def create_cloudformation_stack()-> Tuple[bool, bool]:
    cloudformation_client = get_cloudformation_client()

    # stack name
    cluster_name = get_cluster_name()
    parameters = [{"ParameterKey": "ClusterName", 'ParameterValue': cluster_name}]

    file_name = resource_filename('tensorkube', 'configurations/karpenter_cloudformation.yaml')

    try:
        stack = cloudformation_client.describe_stacks(StackName=cluster_name)
        if stack['Stacks'][0]['StackStatus'] == 'CREATE_COMPLETE':
            click.echo(f"Stack {cluster_name} already exists. Skipping stack creation")
            creation_queued = False
            in_created_state = True
            return creation_queued, in_created_state
        elif stack['Stacks'][0]['StackStatus'] == 'CREATE_IN_PROGRESS':
            click.echo(f"Stack {cluster_name} already exists and is in creation state. Please wait for "
                       f"creation to complete.")
            creation_queued = True
            in_created_state = False
            return creation_queued, in_created_state
        elif stack['Stacks'][0]['StackStatus'] == 'ROLLBACK_COMPLETE':
            click.echo(f"Stack {cluster_name} already exists but is in rollback state. Please delete the stack "
                       f"and recreate it.")
            creation_queued = False
            in_created_state = False
            return creation_queued, in_created_state
        elif stack['Stacks'][0]['StackStatus'] == 'DELETE_COMPLETE':
            # use the configurations/karpenter_cloudformation.yaml in the library
            with open(file_name) as file:
                template = file.read()
            response = cloudformation_client.create_stack(StackName=cluster_name, TemplateBody=template, Parameters=parameters,
                                                          Capabilities=["CAPABILITY_NAMED_IAM"])
            creation_queued = True
            in_created_state = False
            return creation_queued, in_created_state
        else:
            click.echo(f"Stack {cluster_name} already exists but not in created state. Either wait for "
                       f"creation to complete or delete stack to recreate it.")
            creation_queued = False
            in_created_state = False
            return creation_queued, in_created_state
    except botocore.exceptions.ClientError as e:
        if 'does not exist' in str(e):
            # use the configurations/karpenter_cloudformation.yaml in the library
            with open(file_name) as file:
                template = file.read()
            response = cloudformation_client.create_stack(StackName=cluster_name, TemplateBody=template, Parameters=parameters,
                                                          Capabilities=["CAPABILITY_NAMED_IAM"])
            creation_queued = True
            in_created_state = False
            return creation_queued, in_created_state
        else:
            click.echo(e)
            raise Exception('Unable to create cloudformation cluster')


def delete_role_and_attached_policies(iam, role_name):
    # List all attached policies
    attached_policies = iam.list_attached_role_policies(RoleName=role_name)['AttachedPolicies']

    # Detach each policy
    for policy in attached_policies:
        iam.detach_role_policy(RoleName=role_name, PolicyArn=policy['PolicyArn'])
        if get_cluster_name() in policy['PolicyName']:
            # Delete the policy
            iam.delete_policy(PolicyArn=policy['PolicyArn'])

    # Delete the role
    iam.delete_role(RoleName=role_name)


def delete_cloudformation_stack(stack_name):
    cloudformation_client = get_cloudformation_client()

    try:
        cloudformation_client.describe_stacks(StackName=stack_name)
    except botocore.exceptions.ClientError as e:
        if 'does not exist' in str(e):
            print(f'Stack {stack_name} does not exist.')
            return
        else:
            raise
    iam = get_iam_client()

    role_name = f'KarpenterNodeRole-{get_cluster_name()}'

    # List all instance profiles
    instance_profiles = iam.list_instance_profiles()['InstanceProfiles']

    # For each instance profile
    for profile in instance_profiles:
        # Check if the role is associated with the instance profile
        for role in profile['Roles']:
            if role['RoleName'] == role_name:
                # Remove the role from the instance profile
                iam.remove_role_from_instance_profile(InstanceProfileName=profile['InstanceProfileName'],
                    RoleName=role_name)

                # Delete the instance profile
                iam.delete_instance_profile(InstanceProfileName=profile['InstanceProfileName'])

    # Delete the role
    delete_role_and_attached_policies(iam, role_name)

    response = cloudformation_client.delete_stack(StackName=stack_name)
    # Create a waiter to wait for the stack to be deleted
    waiter = cloudformation_client.get_waiter('stack_delete_complete')

    # Start streaming the stack events in a separate thread
    threading.Thread(target=stream_stack_events, args=(stack_name,)).start()

    # Wait for the stack to be deleted
    waiter.wait(StackName=stack_name, WaiterConfig={'MaxAttempts': 30})
    return response


def delete_launch_templates():
    ec2_client = get_ec2_client()
    cluster_name = get_cluster_name()
    # Describe the launch templates
    response = ec2_client.describe_launch_templates(
        Filters=[{'Name': 'tag:karpenter.k8s.aws/cluster', 'Values': [cluster_name]}])
    launch_template_names = [lt['LaunchTemplateName'] for lt in response['LaunchTemplates']]

    # Delete each launch template
    for name in launch_template_names:
        ec2_client.delete_launch_template(LaunchTemplateName=name)


def stream_stack_events(stack_name):
    seen_events = set()
    cf_client = get_cloudformation_client()
    while True:
        try:
            events = cf_client.describe_stack_events(StackName=stack_name)['StackEvents']
            for event in reversed(events):
                event_id = event['EventId']
                if event_id not in seen_events:
                    seen_events.add(event_id)
                    click.echo(
                        f"{event['Timestamp']} {event['ResourceStatus']} {event['ResourceType']} {event['LogicalResourceId']} {event.get('ResourceStatusReason', '')}")
            # Check if stack creation is complete
            stack_status = cf_client.describe_stacks(StackName=stack_name)['Stacks'][0]['StackStatus']
            if stack_status.endswith('_COMPLETE') or stack_status.endswith('_FAILED'):
                break
            time.sleep(5)
        except botocore.exceptions.ClientError as e:
            if 'Stack with id' in str(e) and 'does not exist' in str(e):
                break


def cloudformation():
    """Create a cloudformation stack."""
    click.echo("Creating cloudformation stack...")
    # create_cloudformation_stack()
    creation_queued, in_created_state = create_cloudformation_stack()
    if creation_queued:
        stream_stack_events(get_cluster_name())
    click.echo("Cloudformation stack created.")



from enum import Enum

REGION = 'us-east-1'
NAMESPACE = 'kube-system'
DEFAULT_NAMESPACE = 'default'
SERVICE_ACCOUNT_NAME = 's3-csi-driver-sa'
ADDON_NAME = "aws-mountpoint-s3-csi-driver"
KNATIVE_SERVING_NAMESPACE = 'knative-serving'
CONFIG_FEATURES = 'config-features'
BUILD_TOOL = 'buildkit'

def get_cluster_name():
    return "tensorkube"


def get_mount_policy_name(cluster_name):
    return f'{cluster_name}-mountpoint-policy'


def get_mount_driver_role_name(cluster_name):
    return f'{cluster_name}-mountpoint-driver-role'


class Events(Enum):
    CONFIGURE_START = 'configure-start'
    CONFIGURE_END = 'configure-end'
    CONFIGURE_ERROR = 'configure-error'

    DEPLOY_START = 'deploy-start'
    DEPLOY_END = 'deploy-end'
    DEPLOY_ERROR = 'deploy-error'

    TEST_START = 'test-start'
    TEST_END = 'test-end'


class PodStatus(Enum):
    PENDING = 'Pending'
    RUNNING = 'Running'
    SUCCEEDED = 'Succeeded'
    FAILED = 'Failed'
    UNKNOWN = 'Unknown'


def get_efs_service_account_name() -> str:
    return f'{get_cluster_name()}-efs-csi-controller-sa'


def get_efs_role_name() -> str:
    return f'AmazonEKS_EFS_CSI_DriverRole_{get_cluster_name()}'


def get_efs_security_group_name() -> str:
    return f'eks-efs-sg-{get_cluster_name()}'



from tensorkube.services.s3_service import create_s3_bucket
from tensorkube.services.iam_service import create_mountpoint_iam_policy
from tensorkube.services.eks_service import create_eks_addon
from tensorkube.services.k8s_service import create_namespace, create_docker_registry_secret, apply_k8s_buildkit_config
from tensorkube.helpers import create_mountpoint_driver_role_with_policy, get_base64_encoded_docker_config
from tensorkube.constants import NAMESPACE, SERVICE_ACCOUNT_NAME
import os

account_no = "089962644720"
namespace = NAMESPACE
cluster_name = "shared-cluster-6"
region = "us-east-1"
bucket_name = "16114020-test-bucket-6"
mountpoint_policy_name = "16114020-mountpoint-test-policy-6"
mountpoint_driver_role_name = "16114020-mountpoint-test-role-6"
addon_name = "aws-mountpoint-s3-csi-driver"
service_account_name = SERVICE_ACCOUNT_NAME
args = [ "--context=dir:///data/fastapi_app",
            "--dockerfile=/data/fastapi_app/Dockerfile",
            "--destination=divtf/kaniko-test-1:v5"]

if namespace not in ["default", "kube-system"]:
    create_namespace(namespace)

# # NOTE: bucket name must be universally unique
# # NOTE: bucket name should not be guessable
print(create_s3_bucket(bucket_name, region))

# print(upload_folder_to_bucket(bucket_name,
#                               "/Users/divys/Documents/tensorfuse-all/testing/fastapi_app/",
#                               region=region,
#                               s3_path="fastapi_app/"))


print(create_mountpoint_iam_policy(mountpoint_policy_name, bucket_name, region=region))


print(create_mountpoint_driver_role_with_policy(cluster_name, account_no, mountpoint_driver_role_name, mountpoint_policy_name, service_account_name, namespace, region))

print(create_eks_addon(cluster_name, addon_name, account_no, mountpoint_driver_role_name, region))

DOCKER_USERNAME = os.evniron.get("DOCKER_USERNAME")
DOCKER_PASSWORD = os.evniron.get("DOCKER_PASSWORD")
DOCKER_EMAIL = os.evniron.get("DOCKER_EMAIL")
base64_encoded_docker_config = get_base64_encoded_docker_config(DOCKER_USERNAME, DOCKER_PASSWORD, DOCKER_EMAIL)

create_docker_registry_secret("dockerhub-creds", namespace, base64_encoded_docker_config)

# apply_k8s_buildkit_config("test", "test", bucket_name, get_aws_account_id(), ecr_repo_url)



import platform
import subprocess
import os
import click

from tensorkube.configurations.configuration_urls import DOMAIN_SERVER_URL, KNATIVE_ISTIO_CONTROLLER_URL
from tensorkube.services.eks_service import get_pods_using_namespace, apply_yaml_from_url, delete_resources_from_url


def check_and_install_istioctl():
    """Check if istioctl is installed and install it if it's not."""
    try:
        subprocess.run(["istioctl", "version"], check=True)
        print("istioctl is already installed.")
    except (subprocess.CalledProcessError, FileNotFoundError):
        print("istioctl is not installed. Proceeding with installation.")
        if platform.system() == "Darwin":
            try:
                subprocess.run(["brew", "install", "istioctl"], check=True)
            except subprocess.CalledProcessError as e:
                print("Unable to install istioctl using Homebrew. Please install istioctl manually.")
                raise e
        elif platform.system() == "Linux":
            try:
                install_command = "curl -sL https://istio.io/downloadIstioctl | sh -"
                # Download and install istioctl
                subprocess.run(install_command, shell=True,
                               check=True)

                # Add istioctl to PATH
                istioctl_path = os.path.expanduser("~/.istioctl/bin")
                if istioctl_path not in os.environ["PATH"]:
                    os.environ["PATH"] += os.pathsep + istioctl_path
                    # Optionally, you can add this path to .bashrc or .bash_profile to make it permanent
                    with open(os.path.expanduser("~/.bashrc"), "a") as bashrc:
                        bashrc.write(f'\nexport PATH="$HOME/.istioctl/bin:$PATH"\n')

                print("istioctl installed successfully and PATH updated.")
            except subprocess.CalledProcessError as e:
                print("Unable to install istioctl using curl. Please install istioctl manually.")
                raise e
        else:
            print("Unsupported operating system. Please install istioctl manually.")
            raise Exception('Unsupported operating system.')

        # Verify istioctl installation
        try:
            subprocess.run(["istioctl", "version"], check=True)
            print("istioctl installed successfully.")
        except subprocess.CalledProcessError as e:
            print("istioctl installation failed. Please install istioctl manually.")
            raise e


def install_istio_on_cluster():
    """Install Istio with the default profile."""
    try:
        subprocess.run(["istioctl", "install", "--set", "profile=default", "-y"])
        print("Istio installed successfully.")
    except Exception as e:
        print(f"Error installing Istio: {e}")
        raise e
    # finally using the kubeconfi
    pods = get_pods_using_namespace("istio-system")
    for pod in pods.items:
        click.echo(f"Pod name: {pod.metadata.name}, Pod status: {pod.status.phase}")


def remove_domain_server():
    delete_resources_from_url(DOMAIN_SERVER_URL, "removing Knative Default Domain")


def uninstall_istio_from_cluster():
    """Uninstall Istio from the cluster."""
    # remove knative istion controller
    delete_resources_from_url(KNATIVE_ISTIO_CONTROLLER_URL, "uninstalling Knative Net Istio")
    # remove istio
    try:
        subprocess.run(["istioctl", "x", "uninstall", "--purge", "-y"])
        click.echo("Istio uninstalled successfully.")
    except Exception as e:
        click.echo(f"Error uninstalling Istio: {e}")


def install_net_istio():
    apply_yaml_from_url(KNATIVE_ISTIO_CONTROLLER_URL, "installing Knative Net Istio")


def install_default_domain():
    apply_yaml_from_url(DOMAIN_SERVER_URL, "installing Knative Default Domain")



import os
import platform
import subprocess
from string import Template

from pkg_resources import resource_filename

from tensorkube.services.aws_service import get_aws_account_id, get_karpenter_namespace
from tensorkube.constants import get_cluster_name


# create base cluster using eksctl

def create_base_tensorkube_cluster_eksctl(cluster_name):
    yaml_file_path = resource_filename('tensorkube', 'configurations/base_cluster.yaml')
    # variables
    variables = {"CLUSTER_NAME": cluster_name, "AWS_DEFAULT_REGION": "us-east-1", "K8S_VERSION": "1.29",
                 "AWS_ACCOUNT_ID": get_aws_account_id(), "KARPENTER_NAMESPACE": get_karpenter_namespace(),
                 "AWS_PARTITION": "aws", }

    with open(yaml_file_path) as file:
        template = file.read()
    yaml_content = Template(template).substitute(variables)

    temp_yaml_file_path = "/tmp/temp_cluster.yaml"
    with open(temp_yaml_file_path, 'w') as file:
        file.write(yaml_content)

    # Check if the cluster already exists
    try:
        subprocess.run(["eksctl", "get", "cluster", cluster_name], check=True)
        print(f'Cluster {cluster_name} already exists.')
        os.remove(temp_yaml_file_path)
        return None
    except subprocess.CalledProcessError:
        pass  # Cluster does not exist, we can create it

        # Run the eksctl create cluster command
    command = ["eksctl", "create", "cluster", "-f", temp_yaml_file_path]
    subprocess.run(command, check=True)

    # Remove the temporary file
    os.remove(temp_yaml_file_path)


def delete_cluster():
    command = ["eksctl", "delete", "cluster", "--name", get_cluster_name()]
    subprocess.run(command)


def check_and_install_eksctl():
    """Check if eksctl is installed and if not isntall it."""
    try:
        subprocess.run(["eksctl", "version"], check=True)
    except Exception as e:
        # check if the operating system is mac and install eksctl
        if platform.system() == "Darwin":
            try:
                subprocess.run(["brew", "install", "eksctl"])
            except Exception as mac_e:
                print("Unable to install eksctl. Please install eksctl manually.")
                raise mac_e
        elif platform.system() == "Linux":
            try:
                # curl --silent --location "https://github.com/weaveworks/eksctl/releases/latest/download/eksctl_$(uname -s)_amd64.tar.gz" | tar xz -C /tmp
                # sudo mv /tmp/eksctl /usr/local/bin
                # eksctl version
                download_command = [
                    "curl", "--silent", "--location",
                    "https://github.com/weaveworks/eksctl/releases/latest/download/eksctl_$(uname -s)_amd64.tar.gz",
                    "-o", "/tmp/eksctl.tar.gz"
                ]
                subprocess.run(" ".join(download_command), shell=True, check=True)
                # Extract the tar.gz file
                extract_command = ["tar", "xzf", "/tmp/eksctl.tar.gz", "-C", "/tmp"]
                subprocess.run(extract_command, check=True)
                # Move the binary to /usr/local/bin
                move_command = ["sudo", "mv", "/tmp/eksctl", "/usr/local/bin"]
                subprocess.run(move_command, check=True)
                subprocess.run(["eksctl", "version"])
            except Exception as linux_e:
                print("Unable to install eksctl. Please install eksctl manually.")
                raise linux_e
        else:
            print("eksctl is not installed. Please install eksctl manually.")
            raise e



import time
from threading import Thread
from typing import List

import click
import yaml
from kubernetes import client, config, utils, watch
from kubernetes.client import ApiTypeError, ApiException
from pkg_resources import resource_filename
from tensorkube.constants import get_cluster_name

from tensorkube.constants import NAMESPACE, REGION, PodStatus, BUILD_TOOL


def create_namespace(namespace_name):
    config.load_kube_config()
    namespace = client.V1Namespace()
    namespace.metadata = client.V1ObjectMeta(name=namespace_name)
    v1 = client.CoreV1Api()
    v1.create_namespace(body=namespace)


def create_docker_registry_secret(secret_name: str, namespace: str, base64_encoded_dockerconfigjson: str):
    config.load_kube_config()
    v1 = client.CoreV1Api()

    secret = client.V1Secret()
    secret.api_version = "v1"
    secret.kind = "Secret"
    secret.metadata = client.V1ObjectMeta(name=secret_name, namespace=namespace)
    secret.type = "kubernetes.io/dockerconfigjson"
    secret.data = {".dockerconfigjson": base64_encoded_dockerconfigjson}

    v1.create_namespaced_secret(namespace=namespace, body=secret)


def create_aws_secret(credentials, namespace: str = NAMESPACE):
    config.load_kube_config()
    v1 = client.CoreV1Api()
    secret_name = "aws-secret"

    secret = client.V1Secret()
    secret.metadata = client.V1ObjectMeta(name=secret_name)
    secret.string_data = {"AWS_ACCESS_KEY_ID": credentials.access_key, "AWS_SECRET_ACCESS_KEY": credentials.secret_key,
                          "AWS_SESSION_TOKEN": credentials.token}

    try:
        # Check if the secret already exists
        existing_secret = v1.read_namespaced_secret(name=secret_name, namespace=namespace)
        # If the secret exists, update it
        v1.replace_namespaced_secret(name=secret_name, namespace=namespace, body=secret)
        print(f"Secret {secret_name} updated successfully in namespace {namespace}.")
    except ApiException as e:
        if e.status == 404:
            # Secret does not exist, create it
            v1.create_namespaced_secret(namespace=namespace, body=secret)
            print(f"Secret {secret_name} created successfully in namespace {namespace}.")
        else:
            print(f"An error occurred: {e}")
            raise e


def delete_aws_secret(namespace: str = NAMESPACE):
    config.load_kube_config()
    v1 = client.CoreV1Api()
    try:
        v1.read_namespaced_secret(name="aws-secret", namespace=namespace)
    except client.ApiException as e:
        if e.status == 404:
            return
        else:
            raise
    v1.delete_namespaced_secret(name="aws-secret", namespace=namespace)


def create_build_pv_and_pvc(bucket_name: str, namespace: str = NAMESPACE, region: str = REGION):
    config.load_kube_config()

    pv_config_file_path = resource_filename('tensorkube', 'configurations/build_configs/pv.yaml')
    pvc_config_file_path = resource_filename('tensorkube', 'configurations/build_configs/pvc.yaml')
    with open(pv_config_file_path) as f:
        pv = yaml.safe_load(f)
    with open(pvc_config_file_path) as f:
        pvc = yaml.safe_load(f)

    pv['spec']['mountOptions'] = ["allow-delete", "region {}".format(region)]
    pv['spec']['csi']['volumeAttributes']['bucketName'] = bucket_name
    pv['metadata']['namespace'] = namespace

    pvc['metadata']['namespace'] = namespace

    k8s_client = client.ApiClient()
    v1 = client.CoreV1Api()

    pv_name = pv['metadata']['name']
    pvc_name = pvc['metadata']['name']

    try:
        # Check if the PV already exists
        v1.read_persistent_volume(name=pv_name)
        print(f"PersistentVolume {pv_name} already exists. Skipping creation.")
    except ApiException as e:
        if e.status == 404:
            utils.create_from_dict(k8s_client, pv)
            print(f"PersistentVolume {pv_name} created successfully.")
        else:
            print(f"An error occurred while checking PersistentVolume: {e}")
            raise e

    try:
        # Check if the PVC already exists
        v1.read_namespaced_persistent_volume_claim(name=pvc_name, namespace=namespace)
        print(f"PersistentVolumeClaim {pvc_name} already exists in namespace {namespace}. Skipping creation.")
    except ApiException as e:
        if e.status == 404:
            # PVC does not exist, proceed to create
            k8s_client = client.ApiClient()
            utils.create_from_dict(k8s_client, pvc)
            print(f"PersistentVolumeClaim {pvc_name} created successfully in namespace {namespace}.")
        else:
            print(f"An error occurred while checking PersistentVolumeClaim: {e}")
            raise e


def apply_k8s_buildkit_config(project_name: str, sanitised_project_name: str, bucket_name: str,
                              aws_account_number: str, ecr_repo_url: str,
                              namespace: str = NAMESPACE, region: str = REGION):
    config.load_kube_config()

    buildkit_config_file_path = resource_filename('tensorkube',
                                                    'configurations/build_configs/buildkit.yaml')
    with open(buildkit_config_file_path) as f:
        buildkit_config = yaml.safe_load(f)
    
    buildkit_config['metadata']['name'] = 'buildkit-{}'.format(sanitised_project_name)
    buildkit_config['metadata']['namespace'] = namespace
    buildkit_config['spec']['template']['spec']['containers'][0]['env'][0]['value'] = region
    buildkit_config['spec']['template']['spec']['containers'][0]['command'] = [ "/bin/sh", "-c",
        f"""
        apk add --no-cache curl unzip aws-cli docker
        chmod 777 /data/images
        mkdir -p /data/cache
        chmod 777 /data/cache
        ls -l /data/images
        aws ecr get-login-password --region {region} | docker login --username AWS --password-stdin {aws_account_number}.dkr.ecr.{region}.amazonaws.com
        buildctl-daemonless.sh build\
            --frontend dockerfile.v0\
            --local context=/data/build/{project_name}\
            --local dockerfile=/data/build/{project_name}\
            --output type=image,name={ecr_repo_url},push=true\
            --export-cache type=s3,region={region},bucket={bucket_name},name={get_cluster_name()}-cache,mode=max\
            --import-cache type=s3,region={region},bucket={bucket_name},name={get_cluster_name()}-cache"""
    ]

    k8s_client = client.ApiClient()
    utils.create_from_dict(k8s_client, buildkit_config)
    click.echo('Deployed a Buildkit image')


def get_build_job_pod_name(sanitised_project_name: str, namespace: str = NAMESPACE):
    config.load_kube_config()
    v1 = client.CoreV1Api()
    pods = v1.list_namespaced_pod(namespace=namespace)
    for pod in pods.items:
        if pod.metadata.name.startswith("{}-{}".format(BUILD_TOOL, sanitised_project_name)):
            return pod.metadata.name
    return None


def check_pod_status(pod_name, namespace):
    # Load kube config
    config.load_kube_config()

    # Create a Kubernetes API client
    v1 = client.CoreV1Api()

    # Get the status of the pod
    pod_status = v1.read_namespaced_pod_status(name=pod_name, namespace=namespace)

    # Return the status of the pod
    return pod_status.status.phase


def find_and_delete_old_build_job(job_name: str, namespace: str = NAMESPACE):
    config.load_kube_config()
    v1 = client.BatchV1Api()
    jobs = v1.list_namespaced_job(namespace=namespace)
    for job in jobs.items:
        if job.metadata.name == job_name:
            if any(condition.type in ['Complete', 'Failed'] for condition in job.status.conditions):
                v1.delete_namespaced_job(name=job.metadata.name, namespace=namespace)
                return True
            else:
                return False

    return True


def start_streaming_pod(pod_name, namespace, status=None, container_name=None):
    config.load_kube_config()
    v1 = client.CoreV1Api()
    # Create a stream to the pod
    # Initialize the Watch class
    watch_client = watch.Watch()
    # Stream events until the pod is ready
    print(f"Streaming events for pod {pod_name} in namespace {namespace}")
    for event in watch_client.stream(v1.list_namespaced_event, namespace=namespace):
        pod = v1.read_namespaced_pod(name=pod_name, namespace=namespace)
        if event['object'].involved_object.kind == 'Pod' and event['object'].involved_object.name == pod_name:
            print("Event: %s %s" % (event['type'], event['object'].message))
        if pod.status.phase != 'Pending':
            break

    print(f"Streaming logs for pod {pod_name} in namespace {namespace}")

    try:
        for event in watch_client.stream(v1.read_namespaced_pod_log, name=pod_name, namespace=namespace, follow=True,
                                         container=container_name):
            click.echo(event)
            if status:
                pod = v1.read_namespaced_pod_status(name=pod_name, namespace=namespace)
                if status.value == pod.status.phase:
                    print(f"Pod {pod_name} has reached {status.value} state")
                    return
    except client.ApiException as e:
        if e.status == 404:
            print(f"Pod {pod_name} not found in namespace {namespace}")
        else:
            raise
    except KeyboardInterrupt:
        print("Log streaming stopped by user")
        return
    except ApiTypeError as e:
        print(f"An error occurred while streaming logs for pod {pod_name} in namespace {namespace}")
        print(e)
        return
    except Exception as e:
        print(f"An unexpected error occurred for pod {pod_name} in namespace {namespace}")
        print(e)
        raise


def start_streaming_service(service_name, namespace):
    # Load kube config
    config.load_kube_config()

    # Create a Kubernetes API client
    v1 = client.CoreV1Api()

    # Stream the service status
    try:
        while True:
            pods = v1.list_namespaced_pod(namespace, label_selector=f'serving.knative.dev/service={service_name}')
            if pods.items:
                print(f"Pods scheduled for service {service_name}: {[pod.metadata.name for pod in pods.items]}")
                break
            time.sleep(5)
        # Start streaming the logs from the pod
        for pod in pods.items:
            thread = Thread(target=start_streaming_pod,
                            args=(pod.metadata.name, namespace, PodStatus.RUNNING, 'user-container'))
            thread.start()
    except client.ApiException as e:
        if e.status == 404:
            print(f"Service {service_name} not found in namespace {namespace}")
        else:
            raise
    except KeyboardInterrupt:
        print("Service status streaming stopped by user")
        return
    except Exception as e:
        print("An unexpected error occurred for service status streaming")
        raise



import concurrent.futures
import logging
import threading

import click
import boto3
import os

from tqdm import tqdm
from boto3.s3.transfer import TransferConfig, S3Transfer

from tensorkube.constants import REGION
from multiprocessing import Pool


# NOTE: bucket name must be universally unique
# NOTE: bucket name should not be guessable
def create_s3_bucket(bucket_name, region=REGION):
    try:
        s3_client = boto3.client('s3', region_name=region)
        if region is None or region == REGION:
            s3_client.create_bucket(Bucket=bucket_name)
        else:
            location = {'LocationConstraint': region}
            s3_client.create_bucket(Bucket=bucket_name, CreateBucketConfiguration=location)
    except Exception as e:
        click.echo(f"Error creating bucket {bucket_name} in region {region}: {e}")
        return False

    return True


def list_s3_buckets(region: str=REGION):
    try:
        s3 = boto3.client('s3', region_name=region)
        response = s3.list_buckets()
        return response['Buckets']
    except Exception as e:
        click.echo(f"Error listing buckets in region {region}: {e}")
        return []


# TODO!: add a .tensorignore file to ignore files
# def upload_folder_to_bucket(bucket_name, folder_path, region=REGION, s3_path = ""):
#     s3 = boto3.client('s3', region_name=region)
#     for root, dirs, files in os.walk(folder_path):
#         for file in files:
#             if file == ".DS_Store":
#                 continue
#             local_file = os.path.join(root, file)
#             s3_key = s3_path + os.path.relpath(local_file, folder_path).replace(os.sep, '/')
#             s3.upload_file(local_file, bucket_name, s3_key)
#     return True

# Set up logging for debugging
logging.basicConfig(level=logging.INFO)


def upload_file(file_name, bucket_name, s3_key, s3_client, progress_bar):
    def upload_progress(chunk):
        # logging.debug(f"Uploading chunk of {chunk} bytes for {file_name}")
        progress_bar.update(chunk)

    config = boto3.s3.transfer.TransferConfig(use_threads=False)
    try:
        s3_client.upload_file(file_name, bucket_name, s3_key, Callback=upload_progress, Config=config)
        # logging.info(f"Successfully uploaded {file_name} to {bucket_name}/{s3_key}")
    except Exception as e:
        logging.error(f"Error uploading {file_name}: {e}")
    finally:
        progress_bar.close()


def upload_files_in_parallel(bucket_name, folder_path, region='us-east-1', s3_path=''):
    s3_client = boto3.client('s3', region_name=region)
    files_to_upload = []

    for root, dirs, files in os.walk(folder_path):
        for file in files:
            if file.startswith('.') or file == ".DS_Store":
                continue
            local_file = os.path.join(root, file)
            if '/.' in local_file:
                continue
            files_to_upload.append(local_file)

    # logging.debug(f"Files to upload: {files_to_upload}")
    progress_bars = {}
    locks = {file: threading.Lock() for file in files_to_upload}

    def upload_with_progress(file_name):
        file_size = os.path.getsize(file_name)
        s3_key = s3_path + os.path.relpath(file_name, folder_path).replace(os.sep, '/')
        # logging.debug(f"Uploading {file_name} to {s3_key} in bucket {bucket_name}")
        with locks[file_name]:
            progress_bars[file_name] = tqdm(total=file_size, unit='B', unit_scale=True, desc=s3_key, leave=True)
        upload_file(file_name, bucket_name, s3_key, s3_client, progress_bars[file_name])

    with concurrent.futures.ThreadPoolExecutor() as executor:
        futures = {executor.submit(upload_with_progress, file): file for file in files_to_upload}
        for future in concurrent.futures.as_completed(futures):
            try:
                future.result()
            except Exception as e:
                logging.error(f"Error in future for {futures[future]}: {e}")

    for progress_bar in progress_bars.values():
        progress_bar.close()


def empty_s3_bucket(bucket_name, region=REGION):
    try:
        s3 = boto3.resource('s3', region_name=region)
        bucket = s3.Bucket(bucket_name)
        bucket.objects.all().delete()
    except Exception as e:
        click.echo(f"Error emptying bucket {bucket_name} in region {region}: {e}")
        return False

    return True

def delete_s3_bucket(bucket_name, region=REGION):
    try:
        click.echo(f"Emptying bucket {bucket_name}...")
        empty_s3_bucket(bucket_name, region)
        click.echo(f"Deleting bucket {bucket_name}...")
        s3 = boto3.client('s3', region_name=region)
        s3.delete_bucket(Bucket=bucket_name)
    except Exception as e:
        click.echo(f"Error deleting bucket {bucket_name} in region {region}: {e}")
        return False

    return True
